@@ -21,10 +21,16 @@ import com.android.apksig.SigningCertificateLineage;
 import com.android.apksig.apk.ApkFormatException;
 import com.android.apksig.apk.ApkSigningBlockNotFoundException;
 import com.android.apksig.apk.ApkUtils;
+import com.android.apksig.internal.asn1.Asn1BerParser;
+import com.android.apksig.internal.asn1.Asn1DecodingException;
+import com.android.apksig.internal.asn1.Asn1DerEncoder;
+import com.android.apksig.internal.asn1.Asn1EncodingException;
 import com.android.apksig.internal.util.ByteBufferDataSource;
 import com.android.apksig.internal.util.ChainedDataSource;
 import com.android.apksig.internal.util.Pair;
 import com.android.apksig.internal.util.VerityTreeBuilder;
+import com.android.apksig.internal.x509.RSAPublicKey;
+import com.android.apksig.internal.x509.SubjectPublicKeyInfo;
 import com.android.apksig.internal.zip.ZipUtils;
 import com.android.apksig.util.DataSink;
 import com.android.apksig.util.DataSinks;
@@ -33,6 +39,7 @@ import com.android.apksig.util.DataSources;
 
 import com.android.apksig.util.RunnablesExecutor;
 import java.io.IOException;
+import java.math.BigInteger;
 import java.nio.BufferUnderflowException;
 import java.nio.ByteBuffer;
 import java.nio.ByteOrder;
@@ -58,17 +65,13 @@ import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
 import java.util.Set;
-import java.util.concurrent.ExecutionException;
-import java.util.concurrent.ForkJoinPool;
-import java.util.concurrent.Future;
 import java.util.concurrent.atomic.AtomicInteger;
-import java.util.function.Function;
 import java.util.function.Supplier;
 import java.util.stream.Collectors;
 
 public class ApkSigningBlockUtils {
 
-    private static final char[] HEX_DIGITS = "01234567890abcdef".toCharArray();
+    private static final char[] HEX_DIGITS = "0123456789abcdef".toCharArray();
     private static final long CONTENT_DIGESTED_CHUNK_MAX_SIZE_BYTES = 1024 * 1024;
     public static final int ANDROID_COMMON_PAGE_ALIGNMENT_BYTES = 4096;
     public static final byte[] APK_SIGNING_BLOCK_MAGIC =
@@ -775,6 +778,53 @@ public class ApkSigningBlockUtils {
         byte[] encodedPublicKey = null;
         if ("X.509".equals(publicKey.getFormat())) {
             encodedPublicKey = publicKey.getEncoded();
+            // if the key is an RSA key check for a negative modulus
+            if ("RSA".equals(publicKey.getAlgorithm())) {
+                try {
+                    // Parse the encoded public key into the separate elements of the
+                    // SubjectPublicKeyInfo to obtain the SubjectPublicKey.
+                    ByteBuffer encodedPublicKeyBuffer = ByteBuffer.wrap(encodedPublicKey);
+                    SubjectPublicKeyInfo subjectPublicKeyInfo = Asn1BerParser.parse(
+                            encodedPublicKeyBuffer, SubjectPublicKeyInfo.class);
+                    // The SubjectPublicKey is encoded as a bit string within the
+                    // SubjectPublicKeyInfo. The first byte of the encoding is the number of padding
+                    // bits; store this and decode the rest of the bit string into the RSA modulus
+                    // and exponent.
+                    ByteBuffer subjectPublicKeyBuffer = subjectPublicKeyInfo.subjectPublicKey;
+                    byte padding = subjectPublicKeyBuffer.get();
+                    RSAPublicKey rsaPublicKey = Asn1BerParser.parse(subjectPublicKeyBuffer,
+                            RSAPublicKey.class);
+                    // if the modulus is negative then attempt to reencode it with a leading 0 sign
+                    // byte.
+                    if (rsaPublicKey.modulus.compareTo(BigInteger.ZERO) < 0) {
+                        // A negative modulus indicates the leading bit in the integer is 1. Per
+                        // ASN.1 encoding rules to encode a positive integer with the leading bit
+                        // set to 1 a byte containing all zeros should precede the integer encoding.
+                        byte[] encodedModulus = rsaPublicKey.modulus.toByteArray();
+                        byte[] reencodedModulus = new byte[encodedModulus.length + 1];
+                        reencodedModulus[0] = 0;
+                        System.arraycopy(encodedModulus, 0, reencodedModulus, 1,
+                                encodedModulus.length);
+                        rsaPublicKey.modulus = new BigInteger(reencodedModulus);
+                        // Once the modulus has been corrected reencode the RSAPublicKey, then
+                        // restore the padding value in the bit string and reencode the entire
+                        // SubjectPublicKeyInfo to be returned to the caller.
+                        byte[] reencodedRSAPublicKey = Asn1DerEncoder.encode(rsaPublicKey);
+                        byte[] reencodedSubjectPublicKey =
+                                new byte[reencodedRSAPublicKey.length + 1];
+                        reencodedSubjectPublicKey[0] = padding;
+                        System.arraycopy(reencodedRSAPublicKey, 0, reencodedSubjectPublicKey, 1,
+                                reencodedRSAPublicKey.length);
+                        subjectPublicKeyInfo.subjectPublicKey = ByteBuffer.wrap(
+                                reencodedSubjectPublicKey);
+                        encodedPublicKey = Asn1DerEncoder.encode(subjectPublicKeyInfo);
+                    }
+                } catch (Asn1DecodingException | Asn1EncodingException e) {
+                    System.out.println("Caught a exception encoding the public key: " + e);
+                    e.printStackTrace();
+                    encodedPublicKey = null;
+                }
+            }
         }
         if (encodedPublicKey == null) {
             try {
